import requests
import json
import logging
import time
import os
import contextlib
import uuid
from config import (
    API_KEY, API_SECRET, AUTH_ENDPOINT, BASE_ENDPOINT,
    MESSAGES_ENDPOINT, REFRESH_TOKEN_ENDPOINT, CHATS_ENDPOINT, 
)

logger = logging.getLogger(__name__)

@contextlib.contextmanager
def file_lock(file_path):
    """Windows-specific file locking context manager"""
    lock_file = file_path + ".lock"
    while True:
        try:
            # Try to create lock file
            lock_fd = os.open(lock_file, os.O_CREAT | os.O_EXCL | os.O_RDWR)
            break
        except FileExistsError:
            # If lock exists, wait and retry
            time.sleep(0.1)
    try:
        yield
    finally:
        # Clean up lock file
        os.close(lock_fd)
        try:
            os.unlink(lock_file)
        except OSError:
            pass

class APIService:
    _access_token = None
    _refresh_token = None
    _token_expiry = None
    BASE_ENDPOINT = BASE_ENDPOINT  # Add this line to make BASE_ENDPOINT a class attribute

    @classmethod
    def get_access_token(cls):
        # Check if existing token is still valid
        if cls._access_token and cls._token_expiry and cls._token_expiry > time.time():
            return cls._access_token

        try:
            headers = {
                "api-key": API_KEY,
                "api-secret": API_SECRET
            }
            response = requests.post(AUTH_ENDPOINT, headers=headers)
            
            if response.status_code == 200:
                data = response.json()['data']
                cls._access_token = data['accessToken']
                cls._refresh_token = data['refreshToken']
                cls._token_expiry = time.time() + 3500  # Set expiry to slightly less than 1 hour
                return cls._access_token
            else:
                logger.error(f"Failed to get access token. Status code: {response.status_code}")
        except Exception as e:
            logger.error(f"Error getting access token: {str(e)}", exc_info=True)
        return None

    @classmethod
    def refresh_token(cls):
        if not cls._refresh_token:
            return cls.get_access_token()

        headers = {
            "api-key": API_KEY,
            "Content-Type": "application/json"
        }
        payload = json.dumps({"refreshToken": cls._refresh_token})
        
        try:
            response = requests.post(REFRESH_TOKEN_ENDPOINT, headers=headers, data=payload)
            if response.status_code == 200:
                data = response.json()['data']
                cls._access_token = data['accessToken']
                cls._refresh_token = data['refreshToken']
                return cls._access_token
        except Exception as e:
            logger.error(f"Error refreshing token: {str(e)}", exc_info=True)
        return None

    @classmethod
    def fetch_messages(cls, start_date, end_date):
        logger.info(f"Fetching messages from {start_date} to {end_date}")
        access_token = cls.get_access_token()
        if not access_token:
            logger.error("Could not get access token")
            return []

        headers = {
            "Authorization": f"Bearer {access_token}",
            "Content-Type": "application/json"
        }
        params = {
            "startDate": start_date,
            "endDate": end_date,
        }
        
        all_data = []
        next_url = MESSAGES_ENDPOINT
        page_count = 0
        max_pages = 2  # Limit maximum pages to prevent infinite loops

        while next_url and page_count < max_pages:
            try:
                logger.info(f"Fetching page {page_count + 1} from URL: {next_url}")
                response = requests.get(
                    next_url, 
                    headers=headers, 
                    params=params if next_url == MESSAGES_ENDPOINT else None,
                    timeout=30  # Add timeout
                )

                if response.status_code == 401:
                    logger.warning("Access token expired, attempting to refresh")
                    access_token = cls.refresh_token()
                    if access_token:
                        headers["Authorization"] = f"Bearer {access_token}"
                        response = requests.get(next_url, headers=headers, params=params)
                    else:
                        logger.error("Failed to refresh token")
                        break

                if response.status_code == 200:
                    data = response.json()
                    if not isinstance(data, dict):
                        logger.error(f"Invalid response format: {data}")
                        break

                    items = data.get("data", {}).get("items", [])
                    if not items:
                        logger.info("No more items found")
                        break

                    all_data.extend(items)
                    logger.info(f"Retrieved {len(items)} messages from page {page_count + 1}")

                    # Handle pagination
                    next_page = data.get("data", {}).get("links", {}).get("next")
                    if next_page:
                        next_url = (
                            next_page if next_page.startswith('http') 
                            else f"{BASE_ENDPOINT.rstrip('/')}/{next_page.lstrip('/')}"
                        )
                    else:
                        break
                else:
                    logger.error(f"Error response {response.status_code}: {response.text}")
                    break

            except requests.exceptions.RequestException as e:
                logger.error(f"Request error: {str(e)}")
                break
            except Exception as e:
                logger.error(f"Unexpected error: {str(e)}", exc_info=True)
                break

            page_count += 1

        logger.info(f"Total messages fetched: {len(all_data)}")
        return all_data


    @classmethod
    def fetch_chats(cls, start_date, end_date):
        logger.info(f"Fetching chats from {start_date} to {end_date}")
        access_token = cls.get_access_token()
        if not access_token:
            logger.error("Could not get access token")
            return []

        headers = {"Authorization": f"Bearer {access_token}"}
        params = {
            "startDate": start_date,
            "endDate": end_date
        }

        try:
            logger.debug(f"Making request to {CHATS_ENDPOINT} with params: {params}")
            response = requests.get(CHATS_ENDPOINT, headers=headers, params=params)
            logger.debug(f"Response status: {response.status_code}")

            if response.status_code == 401:
                logger.warning("Access token expired, attempting to refresh")
                access_token = cls.refresh_token()
                if access_token:
                    headers = {"Authorization": f"Bearer {access_token}"}
                    response = requests.get(CHATS_ENDPOINT, headers=headers, params=params)
                else:
                    logger.error("Failed to refresh token")
                    return []

            if response.status_code == 200:
                data = response.json()
                logger.debug(f"Received response data: {data}")
                if "data" in data and "items" in data["data"]:
                    return data["data"]["items"]
                else:
                    logger.error("Invalid response format from chats endpoint")
                    return []
            else:
                logger.error(f"Failed to fetch chats: {response.status_code}")
                return []
            
        except Exception as e:
            logger.error(f"Error fetching chats: {str(e)}", exc_info=True)
            return []

    @classmethod
    def download_media(cls, media_url, save_path):
        """
        Download media file from API with proper error handling and retries
        """
        temp_path = None
        try:
            directory = os.path.dirname(save_path)
            if not os.path.exists(directory):
                os.makedirs(directory, exist_ok=True)

            if os.path.exists(save_path):
                return True

            # Generate unique temp file name
            temp_path = f"{save_path}.{uuid.uuid4().hex}.tmp"
            max_retries = 3
            retry_count = 0

            while retry_count < max_retries:
                try:
                    access_token = cls.get_access_token()
                    if not access_token:
                        logger.error("Failed to get access token for media download")
                        retry_count += 1
                        time.sleep(1)
                        continue

                    full_url = (
                        media_url if media_url.startswith('http')
                        else f"{BASE_ENDPOINT.rstrip('/')}/{media_url.lstrip('/')}"
                    )

                    logger.info(f"Attempting download from URL: {full_url}")
                    headers = {"Authorization": f"Bearer {access_token}"}
                    
                    response = requests.get(full_url, headers=headers, stream=True)
                    if response.status_code == 200:
                        # Write to temp file
                        with open(temp_path, 'wb') as f:
                            for chunk in response.iter_content(chunk_size=8192):
                                if chunk:
                                    f.write(chunk)
                        
                        # Ensure file handle is completely closed
                        time.sleep(0.5)  # Increased delay
                        
                        try:
                            # Check if target file exists
                            if os.path.exists(save_path):
                                os.remove(temp_path)
                                return True
                            
                            # Try to move temp file to final location
                            os.rename(temp_path, save_path)  # Using rename instead of replace
                            time.sleep(0.1)  # Small delay after rename
                            return True
                            
                        except PermissionError:
                            # If rename fails, try alternative approach
                            try:
                                time.sleep(0.5)  # Wait before retry
                                if os.path.exists(temp_path):
                                    with open(temp_path, 'rb') as src:
                                        with open(save_path, 'wb') as dst:
                                            dst.write(src.read())
                                    time.sleep(0.1)  # Wait before cleanup
                                    os.remove(temp_path)
                                return True
                            except Exception as e:
                                logger.error(f"Alternative file move failed: {str(e)}")
                                raise

                    logger.error(f"Download failed with status code: {response.status_code}")
                    if response.status_code == 401:
                        cls._access_token = None

                except requests.exceptions.RequestException as e:
                    logger.error(f"Download attempt {retry_count + 1} failed: {str(e)}")
                except Exception as e:
                    logger.error(f"Unexpected error during download: {str(e)}")

                retry_count += 1
                if retry_count < max_retries:
                    time.sleep(2 ** retry_count)

            return False

        except Exception as e:
            logger.error(f"Unhandled error in download_media: {str(e)}", exc_info=True)
            return False

        finally:
            # Cleanup temp file if it exists
            if temp_path and os.path.exists(temp_path):
                try:
                    time.sleep(0.5)  # Wait before cleanup
                    os.remove(temp_path)
                except Exception as e:
                    logger.error(f"Failed to cleanup temp file {temp_path}: {str(e)}")


def parse_json_string(obj):
    """
    Safely parse JSON strings while leaving non-JSON data untouched.
    Only attempts to parse strings that look like JSON objects or arrays.
    """
    if not isinstance(obj, (str, dict, list)):
        return obj
        
    if isinstance(obj, str):
        # Skip parsing for obvious non-JSON strings
        if not obj or obj.isspace():
            return obj
            
        # Only try to parse if it looks like JSON
        stripped = obj.strip()
        if not (stripped.startswith('{') or stripped.startswith('[')):
            return obj
            
        try:
            return json.loads(obj)
        except json.JSONDecodeError:
            return obj
            
    elif isinstance(obj, dict):
        return {k: parse_json_string(v) for k, v in obj.items()}
        
    elif isinstance(obj, list):
        return [parse_json_string(item) for item in obj]
        
    return obj
